import configparser
import os
from io import StringIO
from json import loads
from pathlib import Path
from typing import Dict, Optional

from configclasses.exceptions import DependencyNotInstalled
from configclasses.helpers import normalize_field_name


def load_env(path: Optional[Path] = None, string: Optional[str] = None):
    try:
        from dotenv import load_dotenv
    except ImportError:
        raise DependencyNotInstalled("You must install 'python-dotenv'")
    if path:
        load_dotenv(dotenv_path=path)
    else:
        load_dotenv(stream=StringIO(string))


def load_dict(dict: Dict[str, str]):
    for k, v in dict.items():
        if isinstance(v, Dict):
            inner_dict = {
                f"{k}_{inner_key}": inner_value for inner_key, inner_value in v.items()
            }
            load_dict(inner_dict)
            continue
        os.environ[normalize_field_name(k)] = str(v)


def load_toml(path: Optional[Path] = None, string: Optional[str] = None):
    try:
        from tomlkit import parse
    except ImportError:
        raise DependencyNotInstalled("You must install 'tomlkit'")

    if path:
        with path.open("r") as config_file:
            cfg = parse(config_file.read())
    else:
        cfg = parse(string)
    load_dict(cfg)


def load_yaml(path: Optional[Path] = None, string: Optional[str] = None):
    try:
        from yaml import full_load
    except ImportError:
        raise DependencyNotInstalled("You must install pyyaml")

    if path:
        with path.open("r") as config_file:
            cfg = full_load(config_file.read())
    else:
        cfg = full_load(string)
    load_dict(cfg)


def load_ini(path: Optional[Path] = None, string: Optional[str] = None):
    cfg = configparser.ConfigParser()
    cfg.read(path) if path else cfg.read_string(string)
    load_dict(cfg.__dict__["_sections"])


def load_json(path: Optional[Path] = None, string: Optional[str] = None):
    if path:
        with path.open("r") as config_file:
            cfg = loads(config_file.read())
    else:
        cfg = loads(string)
    load_dict(cfg)



from dataclasses import _process_class, fields
from pathlib import Path
from typing import Dict, Optional

from configclasses.exceptions import ConfigFilePathDoesNotExist, NonSupportedExtension
from configclasses.helpers import fill_init_dict, supported_extensions
from configclasses.loaders import (
    load_env,
    load_toml,
    load_yaml,
    load_ini,
    load_json,
)


def configclass(
    cls=None, /, *, prefix: Optional[str] = None, **dataclass_parameters,
):
    """Same behaviour that dataclass with additional classmethods as dataclass intializers: from_environ and from_path
    """

    init = dataclass_parameters.get("init", True)
    repr = dataclass_parameters.get("repr", True)
    eq = dataclass_parameters.get("eq", True)
    order = dataclass_parameters.get("order", False)
    unsafe_hash = dataclass_parameters.get("unsafe_hash", False)
    frozen = dataclass_parameters.get("frozen", True)

    def wrap(cls):
        return _post_process_class(
            _process_class(cls, init, repr, eq, order, unsafe_hash, frozen), prefix
        )

    # See if we're being called as @configclass or @configclass().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    return wrap(cls)


def _post_process_class(the_class, the_prefix: Optional[str]):
    def from_environ(
        cls, defaults: Dict[str, str] = None, parent_field_name: Optional[str] = None
    ):
        fields_tuple = [
            (field.name, field.type, field.default) for field in fields(cls)
        ]
        init_dict = fill_init_dict(
            fields_tuple, defaults, parent_field_name, the_prefix
        )
        return cls(**init_dict)

    def from_path(cls, config_path: str, defaults: Dict[str, str] = None):
        path_to_env(Path(config_path))
        return cls.from_environ(defaults)

    def from_string(cls, string: str, extension: str, defaults: Dict[str, str] = None):
        load_file(string=string, extension=extension)
        return cls.from_environ(defaults)

    the_class.from_environ = classmethod(from_environ)
    the_class.from_path = classmethod(from_path)
    the_class.from_string = classmethod(from_string)

    return the_class


def path_to_env(path: Path):
    """Given a path it loads into os.environ all config files found in this path.
    """
    if not path.exists():
        raise ConfigFilePathDoesNotExist(
            f"Config file path '{str(path)}' does not exist"
        )
    if path.is_file():
        load_file(path)
    else:
        load_path(path)


def file_to_env(
    extension: str, path: Optional[Path] = None, string: Optional[str] = None
):
    if extension == ".env":
        load_env(path, string)
    elif extension == ".toml":
        load_toml(path, string)
    elif extension == ".yaml" or extension == ".yml":
        load_yaml(path, string)
    elif extension == ".ini" or extension == ".cfg":
        load_ini(path, string)
    elif extension == ".json":
        load_json(path, string)


def load_path(path: Path):
    for x in path.iterdir():
        path_to_env(x)


def load_file(
    path: Optional[Path] = None,
    string: Optional[str] = None,
    extension: Optional[str] = None,
):
    extension = path.suffix or path.name if path else extension
    if extension in supported_extensions:
        file_to_env(extension, path, string)
    else:
        raise NonSupportedExtension(f"Extension '{extension}' not supported")



import os
from dataclasses import is_dataclass
from typing import Any, Dict, List, Tuple

supported_extensions = (".env", ".toml", ".yaml", ".yml", ".ini", ".cfg", ".json")
converter_types = (int, float)


def get_field_value_from_environ(field_name: Any):
    return os.environ.get(str.upper(field_name)) or os.environ.get(field_name)


def get_default_value(field_name: Any, defaults: Dict[str, str]):
    if defaults:
        return defaults.get(str.upper(field_name)) or defaults.get(field_name)


def fill_init_dict(
    class_fields: List[Tuple[Any, Any, Any]], defaults, parent_field_name, prefix
):
    init_dict = {}
    for class_field_name, class_field_type, class_field_default in class_fields:
        origin_field_name = get_origin_field_name(
            class_field_name, parent_field_name, prefix
        )

        if is_dataclass(class_field_type):
            init_dict[class_field_name] = class_field_type.from_environ(
                defaults, origin_field_name
            )
        elif field_value := get_field_value_from_environ(
            origin_field_name
        ) or get_default_value(origin_field_name, defaults):
            fill_with_environ_or_provided_defaults(
                class_field_name, class_field_type, field_value, init_dict
            )
        else:
            init_dict[class_field_name] = class_field_default

    return init_dict


def fill_with_environ_or_provided_defaults(
    class_field_name, class_field_type, field_value, init_dict
):
    if class_field_type in converter_types:
        init_dict[class_field_name] = class_field_type(field_value)
    elif class_field_type == bool:
        init_dict[class_field_name] = field_value == "True" or field_value == "true"
    else:
        init_dict[class_field_name] = field_value


def get_origin_field_name(class_field_name, parent_field_name, prefix):
    if not prefix and not parent_field_name:
        origin_field_name = class_field_name
    elif parent_field_name:
        origin_field_name = f"{parent_field_name}_{class_field_name}"
    elif prefix:
        origin_field_name = f"{prefix}_{class_field_name}"
    else:
        origin_field_name = f"{prefix}_{parent_field_name}_{class_field_name}"
    return origin_field_name


def normalize_field_name(field_name: str):
    return str.lower(str(field_name))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['configclasses']

package_data = \
{'': ['*']}

extras_require = \
{'dotenv': ['python-dotenv[dotenv]>=0.12,<0.13'],
 'full': ['tomlkit[toml]>=0.5,<0.6',
          'python-dotenv[dotenv]>=0.12,<0.13',
          'pyyaml[yaml]>=5,<6'],
 'toml': ['tomlkit[toml]>=0.5,<0.6'],
 'yaml': ['pyyaml[yaml]>=5,<6']}

setup_kwargs = {
    'name': '12factor-configclasses',
    'version': '0.3.0',
    'description': 'Like dataclasses but for config.',
    'long_description': '# configclasses\n\n![PyPI](https://img.shields.io/pypi/v/12factor-configclasses)\n[![codecov](https://codecov.io/gh/kingoodie/configclasses/branch/master/graph/badge.svg)](https://codecov.io/gh/kingoodie/configclasses)\n<a href="https://codeclimate.com/github/kingoodie/configclasses/maintainability"><img src="https://api.codeclimate.com/v1/badges/9094f65f5caef64fb993/maintainability" /></a>\n\n\nLike dataclasses but for config.\n\nSpecify your config with a class and load it with your env vars or env files.\n\n\n```python\nimport httpx\nfrom configclasses import configclass\nclass UserAPIClient(httpx.AsyncClient):\n    def __init__(self, config: ClientConfig, *args, **kwargs):\n        self.config = config\n        super().__init__(*args, **kwargs)\n\n    async def get_users(self, headers: Optional[Headers] = None) -> Dict[str, Any]:\n        response = await self.get(f"{self.path}/users", auth=headers)\n        response.raise_for_status()\n        return response.json()\n    \n@configclass\nclass ClientConfig:\n    host: str\n    port: int\n\nconfig = ClientConfig.from_path(".env")\nasync with UserAPIClient(config) as client:\n    users = await client.get_users(auth_headers)\n```\n\n## Features\n\n- Fill your configclasses with existent env vars.\n- Define default values in case these variables have no value at all.\n- Load your config files in env vars following [12factor apps](https://12factor.net) recommendations.\n- Support for _.env_, _yaml_, _toml_, _ini_ and _json_.\n- Convert your env vars with specified type in configclass: `int`, `float`, `str` or `bool`.\n- Use nested configclasses to more complex configurations.\n- Specify a prefix with `@configclass(prefix="<PREFIX>")` to append this prefix to your configclass\'  attribute names.\n- Config groups (__TODO__): https://cli.dev/docs/tutorial/config_groups/\n\n## Requirements\n\nPython 3.8+\n\n\n## Installation\n\nDepending on your chosen config file format you can install:\n\n- .env  ->   ```pip install 12factor-configclasses[dotenv]```\n- .yaml ->   ```pip install 12factor-configclasses[yaml]```\n- .toml ->   ```pip install 12factor-configclasses[toml]```\n- .ini  ->   ```pip install 12factor-configclasses```\n- .json ->   ```pip install 12factor-configclasses```\n\nOr install all supported formats with:\n\n    pip install 12factor-configclasses[full]\n    \n## Usage\n\nThere are three ways to use it:\n\n- Loading an .env file:\n\n```.env\n# .env\nHOST=0.0.0.0\nPORT=8000\nDB_URL=sqlite://:memory:\nGENERATE_SCHEMAS=True\nDEBUG=True\nHTTPS_ONLY=False\nGZIP=True\nSENTRY=False\n```\n\n```python\n#config.py\nfrom configclasses import configclass\n\n\n@configclass\nclass DB:\n    user: str\n    password: str\n    url: str\n\n\n@configclass\nclass AppConfig:\n    host: str\n    port: int\n    db: DB\n    generate_schemas: bool\n    debug: bool\n    https_only: bool\n    gzip: bool\n    sentry: bool\n```\n\n```python\n# app.py\nfrom api.config import AppConfig\n\napp_config = AppConfig.from_path(".env")\napp = Starlette(debug=app_config.debug)\n\nif app_config.https_only:\n    app.add_middleware(\n        HTTPSRedirectMiddleware)\nif app_config.gzip:\n    app.add_middleware(GZipMiddleware)\nif app_config.sentry:\n    app.add_middleware(SentryAsgiMiddleware)\n\n...\n\nregister_tortoise(\n    app,\n    db_url=app_config.db.url,\n    modules={"models": ["api.models"]},\n    generate_schemas=app_config.generate_schemas,\n)\n\nif __name__ == "__main__":\n    uvicorn.run(app, host=app_config.host, port=app_config.port)\n```\n\n    \n- Loading predefined environmental variables:\n\nThe same than before, but instead of:\n\n    app_config = AppConfig.from_path(".env")\n    \nYou will do:\n\n    app_config = AppConfig.from_environ()\n    \n- Loading a file from a string:\n\n```python\ntest_env = """HOST=0.0.0.0\nPORT=8000\nDB_URL=sqlite://:memory:\nGENERATE_SCHEMAS=True\nDEBUG=True\nHTTPS_ONLY=False\nGZIP=True\nSENTRY=False"""\napp_config = AppConfig.from_string(test_env, ".env")\n```',
    'author': 'Pablo Cabezas',
    'author_email': 'pabcabsal@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kingoodie/configclasses',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
